from typing import Any

def unique_sorted(values: list[Any]) -> list[Any]:
    return sorted({v for v in values if v})

def unique_w_preserved_insertion_order(values: list[Any]) -> list[Any]:
    return list(dict.fromkeys(values))



from datetime import date, datetime, timedelta
import pandas as pd

MIN_POSSIBLE_DATE, MAX_POSSIBLE_DATE = date(1900, 1, 1), date(2099, 12, 31)
NOW = datetime.now()
TODAY = date.today()
TIME_PERIODS = {
    'Last 30 Days': (TODAY - timedelta(days=30), TODAY),
    'Last Month': ((TODAY.replace(day=1) - timedelta(days=1)).replace(day=1), TODAY.replace(day=1) - timedelta(days=1)),
    'This Year': (date(TODAY.year, 1, 1), TODAY),
    'Last Year': (date(TODAY.year - 1, 1, 1), TODAY.replace(month=1, day=1) - timedelta(days=1)),
    'Last 365 Days': (TODAY - timedelta(days=365), TODAY),
    'All Dates': (MIN_POSSIBLE_DATE, MAX_POSSIBLE_DATE)
}

def to_datetime(d: date) -> datetime:
    return datetime.combine(d, datetime.min.time())

def every_date(dates_and_date_times: list[date | datetime]):
    """For a list of dates/date times, return a sorted list of every date between the earliest and latest
    date in the data, regardless if that date is present in the data.
    Example: ['2024-06-28', '2024-06-30', '2024-06-30'] -> ['2024-06-28', '2024-06-29', '2024-06-30'].
    The incoming list must have at least two elements"""
    if len(dates_and_date_times) < 2:
        raise ValueError('You must provide at least two dates or date-times')
    dates = [d.date() if isinstance(d, datetime) else d for d in dates_and_date_times]
    return [date_ts.date() for date_ts in pd.date_range(start=min(dates), end=max(dates))]


def _diff_count_and_unit(diff: timedelta) -> tuple[int, str]:
    """Return examples: (4, 'year'), (25, 'day'), (0, 'second')"""
    diff_str = str(diff)
    if diff.days < 1:
        hours, minutes, seconds = diff_str.split(':')
        h, m, s = int(hours), int(minutes), int(float(seconds))
        return (h, 'hour') if h > 1 else (m, 'minute') if m > 1 else (s, 'second')
    else:
        days_unit_str, _ = diff_str.split(',')
        days, _ = days_unit_str.split(' ')
        d = int(days)
        return (d // 365, 'year') if d >= 365 else (d // 30, 'month') if d >= 30 else (d, 'day')


def ago(d: date | datetime) -> str:
    """Return examples: '4 years ago', '2 months ago', '20 days ago', 'yesterday'.
    For durations less than one day, if the argument is a date, it will only return 'today' (due to no time info),
    if the arg is a date-time, example returns are: '18 hours ago', '1 minute ago', None
    Because the argument is not a time, it provides less sub-day specificity than ago_from_date_time."""
    # handle date-only items; convert to date-time
    if not isinstance(d, datetime):
        if d > TODAY:
            raise ValueError('You must supply a date/date-time in the past')
        if TODAY == d:
            return 'today'
        d = to_datetime(d)
    diff: timedelta = NOW - d
    if d > NOW:
        raise ValueError('You must supply a date/date-time in the past')
    if diff.days == 1:
        return 'yesterday'
    cnt, unit = _diff_count_and_unit(diff)
    return f'{cnt} {unit} ago' if cnt == 1 else f'{cnt} {unit}s ago'



from dataclasses import dataclass, field
from enum import auto, Enum
from typing import Any, Sequence

class Direction(Enum):
    ROW = auto()
    COLUMN = auto()
    DIAGONAL = auto()


@dataclass
class MatrixSolver:
    m: list[Sequence[Any]]
    row_cnt: int = field(init=False)
    col_cnt: int = field(init=False)
    size: tuple = field(init=False)
    streak_length: int = 2

    def __post_init__(self):
        self.row_cnt = len(self.m)
        self.col_cnt = max([len(r) for r in self.m])
        self.size = (self.row_cnt, self.col_cnt)
        if self.row_cnt < 3 or self.col_cnt < 3:
            raise ValueError('MatrixSolver must be at least 3x3')
        if self.row_cnt != self.col_cnt:
            raise ValueError('MatrixSolver must have identical height & width')
        if not 2 < self.streak_length <= self.row_cnt:
            raise ValueError('MatrixSolver must have a streak between 2 (the default) and its size')

    @property
    def rows(self) -> list[list[Any]]:
        """[(1, 2, 3), (1, 2, 3), (1, 2, 3)] -> [[1, 2, 3], [1, 2, 3], [1, 2, 3]]"""
        return [[c for c in r] for r in self.m]

    @property
    def columns(self) -> list[list[Any]]:
        """[(1, 2, 3), (1, 2, 3), (1, 2, 3)] -> [[1, 1, 1], [2, 2, 2], [3, 3, 3]]"""
        return [[self.m[j][i] for j in range(len(self.m))] for i in range(len(self.m[0]))]

    def _extract_row_chunks(self) -> list[tuple]:
        """[1, 2, 3, 4], streak_length == 3 -> [(1, 2, 3), (2, 3, 4),]"""
        lists = [[tuple(r[i:i + self.streak_length]) for i in range(len(r) - self.streak_length + 1)] for r in self.m]
        return [tup for lst in lists for tup in lst]

    def _extract_col_chunks(self) -> list[tuple]:
        lists = [[tuple(r[i:i+self.streak_length]) for i in range(len(r) - self.streak_length+1)] for r in self.columns]
        return [tup for lst in lists for tup in lst]

    def _extract_diagonal_chunks(self):
        diagonals = []

        # Extract forward diagonals (bottom-left to top-right)
        for start_col in range(self.col_cnt - self.streak_length + 1):
            for start_row in range(self.row_cnt - self.streak_length + 1):
                diag = [self.m[start_row + i][start_col + i] for i in range(self.streak_length)]
                diagonals.append(tuple(diag))

        # Extract backward diagonals (top-left to bottom-right)
        for start_col in range(self.streak_length - 1, self.col_cnt):
            for start_row in range(self.row_cnt - self.streak_length + 1):
                diag = [self.m[start_row + i][start_col - i] for i in range(self.streak_length)]
                diagonals.append(tuple(diag))

        return diagonals

    def get_streaks(self, directions: Direction | list[Direction] = None) -> list[tuple]:
        """Provide an optional list of Directions.  If no Direction is provided, use Row, Column, and Diagonal.
        [[1, 1, 1], [0, 1, 0], [0, 0, 1]] -> [(1, 1, 1), (1, 1, 1)],
        because there's a streak in the 1st row and one diagonal"""
        if not directions:
            directions = [v for k, v in Direction.__members__.items()]
        directions = [directions] if not isinstance(directions, list) else directions

        d = {Direction.ROW: self._extract_row_chunks(),
             Direction.COLUMN: self._extract_col_chunks(),
             Direction.DIAGONAL: self._extract_diagonal_chunks()}

        return [chunk for direction in directions for chunk in d[direction] if len(set(chunk)) == 1]

    def get_streak_cnt_by_value(self, directions: Direction | list[Direction] = None) -> dict[Any: int]:
        """Same as get_streaks, except [(1, 1, 1), (1, 1, 1)] is returned as {1: 2}"""
        if not directions:
            directions = [v for k, v in Direction.__members__.items()]
        directions = [directions] if not isinstance(directions, list) else directions

        streaks = self.get_streaks(directions)
        streaks_by_value = {s[0]: 0 for s in streaks}
        for s in streaks:
            streaks_by_value[s[0]] += 1

        return streaks_by_value



def color_value(value: int | float, colors: tuple[str, str, str] = ('green', 'red', 'white')):
    positive, negative, neutral = colors
    if isinstance(value, (int, float)):
        value_color = negative if value < 0 else positive if value > 0 else neutral
        return value_color



from setuptools import find_packages, setup

setup(
    name="utilnacki",
    version="0.0.1",
    description="A package containing all non-project-specific helpers",
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    long_description="My long description",
    # long_description_content_type="text/markdown",
    url="https://github.com/bernackimark/utilnacki",
    author="Bernacki",
    author_email="bernackimark@gmail.com",
    extras_require={"dev": "twine>=4.0.2"},
    python_requires=">=3.11",
)
